
# =========================================
#       IMPORTS
# --------------------------------------

from os import environ as env

import rootpath

rootpath.append()

from mybad.tests import helper

import mybad

env['COLORS'] = 'true' # lower prio
env['VERBOSE'] = 'true' # lower prio

env['ERROR_COLORS'] = 'false' # higher prio
env['ERROR_VERBOSE'] = 'false' # higher prio


# =========================================
#       TEST
# --------------------------------------

class TestCase(helper.TestCase):

    # NOTE: quite basic test right now, should add more advanced error message formatting assertions

    def test__import(self):
        self.assertModule(mybad)

    def test__instance(self):
        self.assertIsInstance(mybad.Error(), mybad.Error)

    def test_error(self):
        self.assertTrue(hasattr(mybad.Error(), 'error'))

        some_raised_error = TypeError('No good')

        with self.assertNotRaises(Exception):
            error = mybad.Error(some_raised_error)

        self.assertIsInstance(error.error, TypeError)
        self.assertEqual(str(error.error), 'No good')

    def test_id(self):
        self.assertTrue(hasattr(mybad.Error(), 'id'))

        with self.assertNotRaises(Exception):
            error = mybad.Error()

        self.assertIsInstance(error.id, type(None))
        self.assertEqual(error.id, None)

        with self.assertNotRaises(Exception):
            error = mybad.Error(id = 123)

        self.assertIsInstance(error.id, int)
        self.assertEqual(error.id, 123)

        with self.assertNotRaises(Exception):
            error = mybad.Error(id = '123')

        self.assertIsInstance(error.id, str)
        self.assertEqual(error.id, '123')

    def test_key(self):
        self.assertTrue(hasattr(mybad.Error(), 'key'))

        with self.assertNotRaises(Exception):
            error = mybad.Error()

        self.assertIsInstance(error.key, type(None))
        self.assertEqual(error.key, None)

        with self.assertNotRaises(Exception):
            error = mybad.Error(key = 123)

        self.assertIsInstance(error.key, int)
        self.assertEqual(error.key, 123)

        with self.assertNotRaises(Exception):
            error = mybad.Error(key = '123')

        self.assertIsInstance(error.key, str)
        self.assertEqual(error.key, '123')

    def test_code(self):
        self.assertTrue(hasattr(mybad.Error(), 'code'))

        error = mybad.Error()

        self.assertIsInstance(error.code, type(None))
        self.assertEqual(error.code, None)

        with self.assertNotRaises(Exception):
            error = mybad.Error(code = 123)

        self.assertIsInstance(error.code, int)
        self.assertEqual(error.code, 123)

        with self.assertNotRaises(Exception):
            error = mybad.Error(code = '123')

        self.assertIsInstance(error.code, str)
        self.assertEqual(error.code, '123')

    def test_message(self):
        self.assertTrue(hasattr(mybad.Error(), 'message'))

        with self.assertNotRaises(Exception):
            error = mybad.Error()

        self.assertIsInstance(error.message, str)
        self.assertEqual(error.message, 'Unknown')

        with self.assertNotRaises(Exception):
            error = mybad.Error(message = 123)

        self.assertIsInstance(error.message, str)
        self.assertEqual(error.message, '123')

        with self.assertNotRaises(Exception):
            error = mybad.Error(message = '123')

        self.assertIsInstance(error.message, str)
        self.assertEqual(error.message, '123')

    def test_details(self):
        self.assertTrue(hasattr(mybad.Error(), 'details'))

        with self.assertNotRaises(Exception):
            error = mybad.Error()

        self.assertIsInstance(error.details, dict)
        self.assertEqual(error.details, {})

        with self.assertRaises(TypeError):
            error = mybad.Error(details = 123)

        with self.assertRaises(TypeError):
            error = mybad.Error(details = '123')

        with self.assertNotRaises(Exception):
            error = mybad.Error(details = {})

        self.assertIsInstance(error.details, dict)
        self.assertEqual(error.details, {})

        with self.assertNotRaises(Exception):
            error = mybad.Error(details = {
                'foo': 'bar',
                'baz': [1, 2, 3],
            })

        self.assertIsInstance(error.details, dict)
        self.assertEqual(error.details, {
            'foo': 'bar',
            'baz': [1, 2, 3],
        })

    def test_stack(self):
        self.assertTrue(hasattr(mybad.Error(), 'stack'))

        error = mybad.Error()

        self.assertIsInstance(error.stack, list)
        self.assertGreater(len(error.stack), 0)

    def test_stacktrace(self):
        self.assertTrue(hasattr(mybad.Error(), 'stacktrace'))

        with self.assertNotRaises(Exception):
            error = mybad.Error()

        self.assertIsInstance(error.stacktrace, str)
        # self.assertEqual(error.stacktrace, 'NoneType: None')

    def test___str__(self):
        self.assertTrue(hasattr(mybad.Error(), '__str__'))

        self.assertIsInstance(str(mybad.Error()), str)
        self.assertEqual(str(mybad.Error()), 'Unknown')
        self.assertEqual(str(mybad.Error(TypeError('Boo'))), 'Boo')
        self.assertEqual(str(mybad.Error(TypeError('Boo'), message = 'Boo-hoo')), 'Boo-hoo')


# =========================================
#       MAIN
# --------------------------------------

if __name__ == '__main__':
    helper.run(TestCase)




# =========================================
#       IMPORTS
# --------------------------------------

import rootpath

rootpath.append()


# =========================================
#       EXAMPLE
# --------------------------------------

# NOTE: just a very simple maybe non-saying example for now, should add more advanced examples eventually

from mybad import Error

class ToMuchError(Error):
    pass

def print_money(stash):
    try:
        if isinstance(stash, str) and len(stash) > 13:
            raise Exception('Too much money to print: {0}'.format(stash))

        print('PRINT {0}'.format(''.join(stash)))

    except Exception as error:
        raise ToMuchError(error,
            message = "Out of money printing ink...",
            id = hash(stash),
            key = 'too_much',
            code = 400,
            details = dict(
                stash = stash,
            )
        )

amount = ''

for dollar in range(42):
    amount += '$'

    print_money(amount)





# =========================================
#       DEPS
# --------------------------------------

import sys
import os
import inspect
import unittest
import json
import pprint
import types
import logging
import inspecta

from os import path, environ
try:
    # NOTE on `tox` (2/2): ran into issues with `tox` raising `ncurses` error, so disabling colors when running in `tox` for now
    from colour_runner.result import ColourTextTestResult
except:
    pass

from deepdiff import DeepDiff

import rootpath

rootpath.append()

from six import PY2, PY3, string_types

CURRENT_PATH = path.abspath(path.dirname(__file__))
ROOT_PATH = path.abspath(os.getcwd())
# ROOT_PATH = path.abspath(path.join(CURRENT_PATH, '..', '..')) # NOTE: causes trouble in `tox`

print('CURRENT_PATH', CURRENT_PATH)
print('ROOT_PATH', ROOT_PATH)

PACKAGE_SOURCE_DIRECTORY = 'mybad'
PACKAGE_SOURCE_PATH = path.join(ROOT_PATH, PACKAGE_SOURCE_DIRECTORY)

try:
    sys.path.index(PACKAGE_SOURCE_PATH)
except ValueError:
    sys.path.insert(0, PACKAGE_SOURCE_PATH)

TESTS_PATH = CURRENT_PATH
FIXTURES_PATH = path.join(TESTS_PATH, '__fixtures__')

DEFAULT_TEST_ROOT_PATH = path.abspath(path.dirname(__file__))
DEFAULT_TEST_FILE_PATTERN = environ.get('TESTS_PATTERN', 'test_*.py')

PYTHON_VERSION = sys.version_info

print('# =========================================')
print('#       {0}'.format(__name__))
print('# --------------------------------------\n')

print('PYTHON_VERSION: {0}'.format(PYTHON_VERSION))
print('ROOT_PATH: {0}'.format(ROOT_PATH))
print('CURRENT_PATH: {0}'.format(CURRENT_PATH))
print('PACKAGE_SOURCE: {0}'.format(PACKAGE_SOURCE_DIRECTORY))
print('PACKAGE_SOURCE_PATH: {0}'.format(PACKAGE_SOURCE_PATH))
print('TESTS_PATH: {0}'.format(TESTS_PATH))
print('FIXTURES_PATH: {0}'.format(FIXTURES_PATH))
print('DEFAULT_TEST_ROOT_PATH: {0}'.format(DEFAULT_TEST_ROOT_PATH))
print('DEFAULT_TEST_FILE_PATTERN: {0}'.format(DEFAULT_TEST_FILE_PATTERN))

print('\n# --------------------------------------')


# =========================================
#       HELPERS
# --------------------------------------

# See: https://www.pythonsheets.com/notes/python-tests.html

def run(test):
    loader = unittest.TestLoader()

    if isinstance(test, string_types):
        test_path = test

        if path.isfile(test_path):
            test_path = path.dirname(test_path)

        test = suite(path.abspath(test_path))

    if isinstance(test, unittest.TestSuite):
        _suite = test

    else:
        _suite = unittest.TestSuite()

        if isinstance(test, list):
            tests = test

        else:
            tests = [test]

        for test in tests:
            _suite.addTests(loader.loadTestsFromTestCase(test))

    print('')

    # NOTE on `tox` (2/2): ran into issues with `tox` raising `ncurses` error, so disabling colors when running in `tox` for now
    if 'ColourTextTestResult' in globals():
        result = unittest.runner.TextTestRunner(
            verbosity = 2,
            failfast = True,
            resultclass = ColourTextTestResult,
        ).run(_suite)
    else:
        result = unittest.runner.TextTestRunner(
            verbosity = 2,
            failfast = True,
        ).run(_suite)

    succesful = not result.wasSuccessful()
    exit_code = int(succesful)

    sys.exit(exit_code)

def suite(root = DEFAULT_TEST_ROOT_PATH, pattern = DEFAULT_TEST_FILE_PATTERN):
    root = root or DEFAULT_TEST_ROOT_PATH
    pattern = pattern or DEFAULT_TEST_FILE_PATTERN

    suite = unittest.TestSuite()

    for tests in unittest.defaultTestLoader.discover(root, pattern = pattern):
        for test in tests:
            try:
                suite.addTests(test)

            except Exception as error:
                raise ValueError('Failed to load tests - module import issue? {}'.format(tests))

    return suite

def load(root = DEFAULT_TEST_ROOT_PATH, pattern = DEFAULT_TEST_FILE_PATTERN):
    root = root or DEFAULT_TEST_ROOT_PATH
    pattern = pattern or DEFAULT_TEST_FILE_PATTERN

    unittest.defaultTestLoader.discover(root, pattern = pattern)

def deepdiff(a, b, exclude_types = None):
    exclude_types = exclude_types or [logging.Logger]

    return DeepDiff(a, b, ignore_order = True, report_repetition = True, exclude_types = exclude_types)

def pretty(data):
    return inspecta.inspect(data, indent = 4, depth = None)

def root_path(*args):
    return ROOT_PATH

def fixture_path(*args):
    args = list([FIXTURES_PATH]) + list(args)
    path_parts = filter(lambda value: value is not None, [FIXTURES_PATH] + args)
    return path.abspath(path.join(*path_parts))

def fixture_file(relative_file_path, *args, **kvargs):
    fixture_file_path = fixture_path(relative_file_path)

    return open(fixture_file_path, *args, **kvargs)

def fixture(relative_file_path, *args, **kvargs):
    return fixture_file(relative_file_path, *args, **kvargs)

def json_encode(data = None):
    return json.dumps(data, default = lambda x: repr(x)).replace(' ', '')

def assertModule(actual, expression):
    is_string = isinstance(actual.__file__, string_types)

    if not is_string:
        raise AssertionError('module path `{0}` expected to be a string - {1}'.format(actual, expression and '- {0}'.format(expression) or ''))

    module_file_path = actual.__file__

    is_source_file = (module_file_path.find('/{0}'.format(PACKAGE_SOURCE_DIRECTORY)) > -1) # ensure module is loaded from `src`

    if not is_source_file:
        raise AssertionError('module path `{0}` expected to include `{1}` {2}'.format(module_file_path, PACKAGE_SOURCE_DIRECTORY, expression and '- {0}'.format(expression) or ''))

    module_type = type(actual)

    is_module = (module_type == types.ModuleType)

    if not is_module:
        raise AssertionError('module `{0}` expected to include `{1}` {2}'.format(module_file_path, PACKAGE_SOURCE_DIRECTORY, expression and '- {0}'.format(expression) or ''))

def assertDeepEqual(actual, expected, expression = None, exclude_types = None):
    diff = deepdiff(actual, expected, exclude_types = exclude_types)

    if diff != {}:
        raise AssertionError('\n\n%s\nshould deep equal\n\n%s\ndiff:\n\n%s' % (pretty(actual), pretty(expected), pretty(diff)))

    return True

def assertNotDeepEqual(actual, expected, expression = None, exclude_types = None):
    diff = deepdiff(actual, expected, exclude_types = exclude_types)

    if diff == {}:
        raise AssertionError('\n\n%s\nshould deep equal\n\n%s\ndiff:\n\n%s' % (pretty(actual), pretty(expected), pretty(diff)))

    return True

class assertNotRaises(unittest.TestCase):

    def __init__(self, exception, expression):
        self.exception = exception
        self.expression = expression

    def __enter__(self):
        if self.exception is None:
            try:
                self.expression()

            except:
                info = sys.exc_info()

                self.fail('%s raised' % repr(info[0]))

        else:
            self.assertRaises(self.exception, self.expression)

    def __exit__(self, type, value, traceback):
        pass

# See: https://stackoverflow.com/questions/4319825/python-unittest-opposite-of-assertraises
class TestCase(unittest.TestCase):

    maxDiff = None

    def assertNotRaises(self, exception, expression = None):
        return assertNotRaises(exception, expression)

    def assertDeepEqual(self, result, expected, expression = None, exclude_types = None):
        return assertDeepEqual(result, expected, expression)

    def assertNotDeepEqual(self, result, expected, expression = None, exclude_types = None):
        return assertNotDeepEqual(result, expected, expression)

    def assertModule(self, result, expression = None):
        return assertModule(result, expression)




# =========================================
#       IMPORTS
# --------------------------------------

import re
import traceback
import inspect
import inspecta

from termcolor import colored as color

from os import environ as env


# =========================================
#       CONSTANTS
# --------------------------------------

DEFAULT_ERROR_INDENT = 4
DEFAULT_ERROR_DEPTH = None
DEFAULT_ERROR_COLORS = True
DEFAULT_ERROR_VERBOSE = True


# =========================================
#       ERRORS
# --------------------------------------

class Error(Exception):

    def __init__(self,
        error = None,
        id = None,
        key = None,
        code = None,
        message = None,
        details = None,
    ):
        message = message or (error and str(error)) or 'Unknown'
        message = message and str(message)

        stack = inspect.stack()
        # REVIEW: instead serialize `inspect.stack()` into custom pretty string?
        stacktrace = '\n' + traceback.format_exc()
        details = details or {}

        if not isinstance(details, dict):
            raise TypeError('Expected argument {klass}(details = <details>) to be a `{expected_type}`, but was `{actual_type}`.'.format(
                klass = self.__class__.__name__,
                expected_type = type({}),
                actual_type = type(details),
            ))

        self.error = error
        self.id = id
        self.key = key
        self.code = code
        self.message = message
        self.details = details
        self.stack = stack
        self.stacktrace = stacktrace

    def __repr__(self):
        _name = self.__class__.__name__

        return '{} {}'.format(_name, self.__str__())

    def __str__(self):
        colors = env.get('ERROR_COLORS', None)
        colors = colors or env.get('COLORS', None)

        if colors is None:
            colors = DEFAULT_ERROR_COLORS

        colors = re.search(r'^true|1$', str(colors), flags = re.IGNORECASE)

        verbose = env.get('ERROR_VERBOSE', None)
        verbose = verbose or env.get('VERBOSE', None)

        if verbose is None:
            verbose = DEFAULT_ERROR_VERBOSE

        verbose = re.search(r'^true|1$', str(verbose), flags = re.IGNORECASE)

        string = self.inspect(
            colors = colors,
            verbose = verbose
        )

        return string

    def __nonzero__():
        return self.__bool__()

    def __bool__(self):
        return True

    def inspect(self,
        colors = None,
        verbose = None,
    ):
        if self.message is None:
            _message = '<none>'
        else:
            _message = self.message.split(' - Arguments')[0]

        if colors:
            _message = color(_message, 'red')

        if verbose:
            _details = inspecta.inspect(self.details,
                colors = colors,
            )

        else:
            _details = None

        _message = ' - '.join(list(filter(bool, [
            _message,
            _details
        ])))

        return _message

    @staticmethod
    def cast(error):
        if isinstance(error, Error):
            return error

        else:
            return Error(error)

    @staticmethod
    def object(error, attrs):
        if not isinstance(error, Error):
            error = Error(error)

        extended_error = Error.cast(error)

        return dict({
            'type': error.__class__.__name__,
            'id': extended_error.id,
            'code': extended_error.code,
            'key': extended_error.key,
            'message': extended_error.message,
            'details': extended_error.details,
            'stack': extended_error.stack,
        }, **attrs)





# =========================================
#       IMPORTS
# --------------------------------------

import os
import glob
import setuptools

# DISABLED/BUG: this line fails when `pip install totalrecall` but works `pip install .`
# from totalrecall import __version__


# =========================================
#       FUNCTIONS
# --------------------------------------

def find_data_files(data_file_patterns = [], root_path = None):
    root_path = root_path or os.path.abspath(os.path.dirname(__file__))
    data_file_dirs = []

    for root, dirs, files in os.walk(root_path):
        data_file_dirs.append(root)

    data_files = []

    for data_file_dir in data_file_dirs:
        files = []

        for data_file_pattern in data_file_patterns:
            files += glob.glob(os.path.join(data_file_dir, data_file_pattern))

        if not files:
            continue

        target = os.path.join(root_path, data_file_dir)

        data_files.append((target, files))

    return data_files

def get_readme():
    root_path = os.path.abspath(os.path.dirname(__file__))
    readme_file_path = os.path.join(root_path, 'README.md')

    with open(readme_file_path) as file:
        readme = file.read()

    return readme

def get_requirements():
    root_path = os.path.abspath(os.path.dirname(__file__))
    requirements_file_path = os.path.join(root_path, 'requirements.txt')

    with open(requirements_file_path) as file:
        requirements = [requirement.strip() for requirement in file.readlines()]
        requirements = filter(lambda requirement: len(requirement), requirements)
        requirements = list(requirements)

    return requirements


# =========================================
#       MAIN
# --------------------------------------

name = 'mybad'
version = '0.1.1'
description = 'My friendly error base class - for Python.'
keywords = [
    'error',
    'errors',
    'baseclass',
    'meta',
    'debug',
    'inspector',
    'pretty',
    'color',
    'colors',
    'detailed',
    'stacktrace',
]

readme = get_readme()
requirements = get_requirements()
packages = setuptools.find_packages()
data_files = find_data_files(['*.*'], os.path.join(name, 'tests', '__fixtures__'))

config = {
    'name': name,
    'version': version,
    'description': (description),
    'keywords': keywords,
    'author': 'Jonas Grimfelt',
    'author_email': 'grimen@gmail.com',
    'url': 'https://github.com/grimen/python-{name}'.format(name = name),
    'download_url': 'https://github.com/grimen/python-{name}'.format(name = name),
    'project_urls': {
        'repository': 'https://github.com/grimen/python-{name}'.format(name = name),
        'bugs': 'https://github.com/grimen/python-{name}/issues'.format(name = name),
    },
    'license': 'MIT',

    'long_description': readme,
    'long_description_content_type': 'text/markdown',

    'classifiers': [
        'Topic :: Software Development :: Libraries',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],

    'packages': packages,
    'package_dir': {
        'totalrecall': name,
    },
    'package_data': {
        '': [
            'MIT-LICENSE',
            'README.md',
        ],
        'totalrecall': [
            '*.*',
        ],
    },
    'data_files': data_files,
    'include_package_data': True,
    'zip_safe': True,

    'install_requires': requirements,
    'setup_requires': [
        'setuptools_git >= 1.2',
    ],
}

setuptools.setup(**config)
